/*
GUIBEGIN
WINDOW , 62, 149, 337, 41, POPUP|CAPTION|SYSMENU|MINBOX|MAXBOX|THICK, , MIDI DisAssembler
	FONT 8, 400, MS Shell Dlg
	TEXT 5, 29, 327, 8, GROUP, , Message, , Click on "Pick file" to convert a MIDI file to text.
	PUSH 145, 4, 40, 14, TABSTOP, , PickButton, ALT "P", &Pick file
DEND
GUIEND
*/

/* This is a REXX version of the MIDI File Disassembler. It converts
 * a MIDI file into a specially formatted text file. It also uses REXX
 * GUI for the file dialog and main window.
 */

OPTIONS "C_CALL LABELCHECK NOSOURCE"

/* Load REXXGUI.DLL, RXMIDI.DLL, and register all their functions. */
LIBRARY rexxgui, rxmidi

/* Let RxMidi functions raise ERROR condition if they fail. So, we don't have
 * to check the return values from functions such as MIDISetEvent().
 */
MidiErr = "ERROR"
MidiHeading = 1

/* Let Reginald raise SYNTAX condition for any error in a Gui function call. */
GuiErr = "SYNTAX"
GuiHeading = 1

/* Create our window, and activate/show it */
GuiCreateWindow('NORMAL')

/* If the user supplied the text filename when he ran this script, then load/convert it. */
ToFilename = ""
FromFilename = ARG(1)
IF FromFilename \== "" THEN ConvertText()

Again:
DO FOREVER

	GuiGetMsg()

	/* None of our handlers below calls GuiWake(). Also, we have no Child Window Layout
	 * scripts. So we omit further checks of GuiObject/GuiSignal.
	 */

	CATCH NOTREADY
		CONDITION('M')
		SIGNAL Again

	CATCH ERROR
		CONDITION('M')
		SIGNAL Again

	CATCH SYNTAX
		CONDITION('M')
		SIGNAL Again

	FINALLY
		GuiDestroyWindow()
END
RETURN











/* Called by Reginald when user clicks on the "Pick file" button. */

WM_CLICK_PickButton:

/* We call GuiGetMsg() below. If ToFilename isn't "" then, he has
 * clicked the "Pick file" button during a conversion. In that case,
 * we abort the conversion.
 */
IF ToFilename \== "" THEN DO
	ToFilename = ""
	GuiAddCtlText("PickButton", "Aborting...")
	RETURN
END

/* Let the user pick the text file. */
DO
	GuiFile('FromFilename', 'EXISTING', 'Select a MIDI file to convert to text', 'MIDI files (*.mid, *.rmi) | *.mid;*.rmi | All files (*.*) | *.*')

	CATCH SYNTAX
		IF CONDITION('D') \== "CANCEL" THEN CONDITION('M')
		RETURN
END

ConvertText:

/* Create the text file name from the MIDI filename by chopping
 * off any .MID extension and replacing with .TXT.
 */
ToFilename = EDITNAME(FromFileName, "*.txt", "U")

DO
	/* Change "Pick file" to "Loading..." */
	GuiAddCtlText("PickButton", "Loading...")

	/* Load the MIDI file into RAM. */
	MIDIOpenFile(FromFilename)

	/* Open/Overwrite the text file. */
	STREAM(ToFilename, 'C', 'OPEN WRITE REPLACE')

	/* Format the MThd line. */
	LINEOUT(ToFilename, "MThd | Format="||MIDIGetInfo('TYPE')||" | # of Tracks="MIDIGetInfo('TRKS')||" | Division="||MIDIGetInfo('BASE'))

	/* Assume key signature of C. */
	key = 0

	/* Move to the start of the track (select this track as the "currently selected track") */
	DO WHILE MIDITrack("") \== 0

		/* Add a blank line. */
		LINEOUT(ToFilename,"")

		/* Print the track heading. */
		LINEOUT(ToFilename,"Track #"||MIDITrack() - 1 "******************************************")

		/* Don't omit any time components for the first event in this track. */
		Measure = ""
		Beat = ""
		Clock = ""

		/* Check if the CLOSE BOX has been clicked. We do this only at the beginning of
		 * each track being disassembled. We could put it inside the DO WHILE loop below
		 * to allow a quicker response to the user wanting to abort (ie, abort during
		 * writing out a track), but that can slow down the loop too.
		 */
		GuiGetMsg('CLEAR')
		IF ToFilename == "" THEN RETURN		

		/* For each event in this track... */
		DO WHILE MIDIGetEvent() == ""

			/* Format the time. We pad the Measures:Beats:Clocks to fit nicely in a column,
			 * and omit times that are the same as the previous event.
			 */
			Line = ""
			IF Measure \== MIDIEvent.!Measure THEN DO
				Line = LEFT("", 4 - LENGTH(MIDIEvent.!Measure)) || MIDIEvent.!Measure || ':'
				spaces = 0
			END
			ELSE spaces = 5

			IF spaces == 0 | Beat \== MIDIEvent.!Beat THEN DO
				Line = Line || LEFT("", (spaces + 2) - LENGTH(MIDIEvent.!Beat)) || MIDIEvent.!Beat || ':'
				spaces = 0
			END
			ELSE spaces = spaces + 3

			IF spaces == 0 | Clock \== MIDIEvent.!Clock THEN DO
				Line = Line || LEFT("", (spaces + 3) - LENGTH(MIDIEvent.!Clock)) || MIDIEvent.!Clock
				spaces = 0
			END
			ELSE spaces = spaces + 3

			IF spaces \== 0 THEN Line = LEFT("", spaces)

			/* Save these times to compare to the next event's times. */
			Measure = MIDIEvent.!Measure
			Beat = MIDIEvent.!Beat
			Clock = MIDIEvent.!Clock

			/* Get the event name (and save the status). */
			type = MIDIEvent.!Type
			MIDIEvent.!Type = MIDIEventProp()

			/* A MIDI Voice Category message? (These are messages that have a MIDI channel). */
			IF MIDIEvent.!Channel \== "" THEN DO

				/* Format the type and Channel */
				IF MIDIEvent.!Channel < 10 THEN spaces = " "
				ELSE spaces = ""
				Line = Line || " |" || MIDIEvent.!Type || LEFT("", 12 - LENGTH(MIDIEvent.!Type)) || "| chan=" || spaces || MIDIEvent.!Channel || "   | "

				/* Write out the line so far, replacing spaces with TABs. */
				Line = EXPAND(Line, 'T')
				CHAROUT(ToFilename, Line)

				/* Write out this event's remaining information upon one text line. Each
				 * piece of info is separated by a '|' character.
				 */
				SELECT type

					/* On Note or (Off) Note */
					WHEN 144 THEN DO
						IF MIDIEvent.!Data2 \== 0 THEN LINEOUT(ToFilename, "pitch=" || MIDINoteName(key) "| vol=" || MIDIEvent.!Data2)
						ELSE LINEOUT(ToFilename, "pitch=" || MIDINoteName(key))
					END

					/* Off Note */
					WHEN 128 THEN LINEOUT(ToFilename, "pitch=" || MIDINoteName(key) "| vol=" || MIDIEvent.!Data2)

					/* Aftertouch */
					WHEN 160 THEN LINEOUT(ToFilename, "pitch=" || MIDINoteName(key) "| press=" || MIDIEvent.!Data2)

					/* Controller */
					WHEN 176 THEN LINEOUT(ToFilename, "contr=" || MIDICtlName() "| value=" || MIDIEvent.!Data2)

					/* Program */
					WHEN 192 THEN DO
						spaces = ""
						IF MIDIEvent.!Data1 < 100 THEN spaces = " "
						IF MIDIEvent.!Data1 < 10 THEN spaces = "  "
						LINEOUT(ToFilename, "pgm #=" || spaces || MIDIEvent.!Data1 + 1)
					END /* Program */

					/* Poly Press */
					WHEN 208 THEN LINEOUT(ToFilename, "press=" || MIDIEvent.!Data1)

					/* Pitch Wheel */
					WHEN 224 THEN LINEOUT(ToFilename, "bend=" || MIDIEvent.!Data3)

				END /* SELECT */

			END /* Voice messages */

			/* An event other than a MIDI Voice Message. Write out this event's information
			 * upon one text line (except for Sysex, and text type of events). Each piece of
			 * info is separated by a '|' character.
			 */
			ELSE DO

				/* If a Port # event, we're going to substitute a Device Name */
				IF type == 33 THEN DO
					type = 9
					MIDIEvent.!Type = MIDIEventProp(, 9)
					MIDIEvent.!Data1 = MIDIPortName()
				END

				/* Format the type */
				Line = Line || " |" || MIDIEvent.!Type || LEFT("", 12 - LENGTH(MIDIEvent.!Type)) || "| "

				/* Write out the line so far, replacing spaces with TABs */
				Line = EXPAND(Line, 'T')
				CHAROUT(ToFilename, Line)

				/* Write out the remainder of this event, depending upon its type */

				/* Text Meta-events (ie, "Text" to "Device Name"), "Specific", "Unknown Meta" */
				IF (type \== 0 & type < 10) | type == 127 THEN DO

					len = LENGTH(MIDIEvent.!Data1)

					LINEOUT(ToFilename, "")

					/* We need to write out 11 characters per line -- first in hex, and then the actual characters enclosed in <> */
					bytes = 1
					Line = "            "
					Line2 = "<"
					DO total = 1 TO len

						/* Extract the next character */
						chr = SUBSTR(MIDIEvent.!Data1, total, 1)

						/* Convert to hex and append to Line */
						IF bytes \== 1 THEN Line = Line || " "
						Line = Line || "0x" || C2X(chr)

						/* Append character to the end of Line2. If not a printable char, substitute a '.' */
						IF chr <<= '19'X | char >> '7F'X THEN chr = '.'
						Line2 = Line2 || chr

						/* Increment count of chars on this line */
						bytes = bytes + 1

						/* Time to write out this line (ie, 11 chars formatted)? */
						IF bytes > 11 THEN DO
							LINEOUT(ToFilename, EXPAND(Line || LEFT("", 67 - LENGTH(Line)) || Line2 || ">", 'T'))
							bytes = 1
							Line = "            "
							Line2 = "<"
						END
					END

					/* Write out any partial, final line */
					IF bytes \== 1 THEN LINEOUT(ToFilename, EXPAND(Line || LEFT("", 67 - LENGTH(Line)) || Line2 || ">", 'T'))

				END /* Text Meta-events */

				/* Sysex */
				/* First Packet */
				/* Packet */
				/* Last Packet */
				ELSE IF type == 240 | type == 247 | type == 255 THEN DO

					LINEOUT(ToFilename, "")

					/* We want to write out 11 bytes per line */
					total = 0
					DO WHILE total < MIDIEvent.!Data1

						/* Retrieve the next 11 bytes, formatted C style hex */
						chr = MIDISysex(11, total + 1, 'C')

						/* Write the line */
						LINEOUT(ToFilename, "            " || chr)

						/* Next 'start' */
						total = total + 11
					END

				END /* Sysex */

				ELSE SELECT type

					/* Seq # */
					WHEN 0 THEN LINEOUT(ToFilename, MIDIEvent.!Data1)

					/* Chan # */
					WHEN 32 THEN LINEOUT(ToFilename, MIDIEvent.!Data1)

					/* Tempo */
					WHEN 81 THEN LINEOUT(ToFilename, "BPM=" || MIDIEvent.!Data2 || LEFT("           ", 6 - LENGTH(MIDIEvent.!Data2)) || "| micros\quarter=" || MIDIEvent.!Data1)

					/* Time Sig */
					WHEN 88 THEN DO
						spaces = " "
						pad = 1
						IF LENGTH(MIDIEvent.!Data1) > 1 THEN DO
							spaces = ""
							pad = 0
						END
						LINEOUT(ToFilename, spaces || MIDIEvent.!Data1 || "/" || MIDIEvent.!Data2 || LEFT("           ", 9 - (LENGTH(MIDIEvent.!Data1) + LENGTH(MIDIEvent.!Data2) + pad)) || "| MIDI-clocks\click=" || MIDIEvent.!Data3 "| 32nds\quarter=" || MIDIEvent.!Data4)
					END

					/* SMPTE */
					WHEN 84 THEN LINEOUT(ToFilename, "hour=" || MIDIEvent.!Data1 "| min=" || MIDIEvent.!Data2 "| sec=" || MIDIEvent.!Data3 "| frame=" || MIDIEvent.!Data4 "| subs=" || MIDIEvent.!Data5)

					/* Key Sig */
					WHEN 89 THEN DO
						/* Save the sharps/flats so that we can pass this to MIDINoteName() */
						key = MIDIEvent.!Data1
						IF key = 0 & MIDIEvent.!Data2 == 1 THEN key = -1
						LINEOUT(ToFilename, MIDIEvent.!Data3)
					END

					/* MTC Quarter Frame */
					WHEN 241 THEN LINEOUT(ToFilename, MIDIEvent.!Data1)

					/* Song Position Pointer */
					WHEN 242 THEN LINEOUT(ToFilename, MIDIEvent.!Data3)

					/* Song Select */
					WHEN 243 THEN LINEOUT(ToFilename, MIDIEvent.!Data1)

					/* End of Track */
					/* Tune Request */
					/* MIDI Clock */
					/* MIDI Start */
					/* MIDI Stop */
					/* MIDI Continue */
					/* Active Sense */
					/* These all have no additional fields to write out */
					OTHERWISE LINEOUT(ToFilename, "")

				END /* SELECT */

			END /* Non-Voice message. */

		END /* DO WHILE MIDIGetEvent. */

	END /* DO UNTIL MIDITrack. */

	FINALLY
		/* Close the text file. */
		STREAM(ToFilename, 'C', 'CLOSE')

		/* Change back to "Pick file". */
		GuiAddCtlText("PickButton", "Pick file")

		/* Indicate that the disassembly is done. */
		GuiAddCtlText('Message', 'Done converting' TRANSLATE(ToFileName, "/", "\"))

		/* No longer converting a file. */
		ToFileName = ""
END

RETURN